(******************************************************************************)
(*                                                                            *)
(*                               PROJECT HELM                                 *)
(*                                                                            *)
(*                Claudio Sacerdoti Coen <sacerdot@cs.unibo.it>               *)
(*                                 24/01/2000                                 *)
(*                                                                            *)
(* This module implements a trival cache system (an hash-table) for cic       *)(* objects. Uses the getter (getter.ml) and the parser (cicParser.ml)         *)(*                                                                            *)
(******************************************************************************)

exception CircularDependency of string;;

(* get_obj uri                                                                *)
(* returns the cic object whose uri is uri. If the term is not just in cache, *)
(* then it is parsed via CicParser.term_of_xml from the file whose name is    *)
(* the result of Getter.get uri                                               *)
val get_obj : UriManager.uri -> Cic.obj

(* get_annobj uri                                                             *)
(* returns the cic object whose uri is uri either in annotated and in         *)
(* deannotated form. It returns also the map from ids to annotation targets.  *)
(* The term is put in cache if it's not there yet.                            *)
(* The functions raise CircularDependency if asked to retrieve a Frozen object*)
val get_annobj :
 UriManager.uri -> Cic.annobj * (Cic.id, Cic.anntarget) Hashtbl.t * Cic.obj

(* DANGEROUS!!!                                *)
(* USEFUL ONLY DURING THE FIXING OF THE FILES  *)
(* change_obj uri (Some newobj)                *)
(*  maps uri to newobj in cache.               *)
(* change_obj uri None                         *)
(*  maps uri to a freeze dummy-object.         *)
val change_obj : UriManager.uri -> Cic.obj option -> unit

type type_checked_obj =
   CheckedObj of Cic.obj    (* cooked obj *)
 | UncheckedObj of Cic.obj  (* uncooked obj *)

(* is_type_checked uri cookingsno                                   *)
(*CSC commento falso ed obsoleto *)
(* returns (true,object) if the object has been type-checked        *)
(* otherwise it returns (false,object) and freeze the object for    *)
(* type-checking                                                    *)
(* set_type_checking_info must be called to unfreeze the object     *)
val is_type_checked : UriManager.uri -> int -> type_checked_obj

(* set_type_checking_info uri                                         *)
(* must be called once the type-checking of uri is finished           *)
(* The object whose uri is uri is unfreezed and won't be type-checked *)
(* again in the future (is_type_checked will return true)             *)
val set_type_checking_info : UriManager.uri -> unit

(* get_cooked_obj uri cookingsno *)
val get_cooked_obj : UriManager.uri -> int -> Cic.obj

val cook_obj : (Cic.obj -> UriManager.uri -> (int * Cic.obj) list) ref
